/**
 * Copyright (c) 2023 Eclipse Foundation
 *
 * This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 2.0
 * which is available at https://www.eclipse.org/legal/epl-2.0/
 *
 * Author: Martin Lowe <martin.lowe@eclipse-foundation.org>
 *
 * SPDX-License-Identifier: EPL-2.0
 */
package org.eclipsefoundation.git.eca.tasks;

import org.eclipse.microprofile.config.inject.ConfigProperty;
import org.eclipsefoundation.git.eca.helper.GithubHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import io.quarkus.scheduler.Scheduled;
import jakarta.annotation.PostConstruct;
import jakarta.enterprise.context.ApplicationScoped;
import jakarta.enterprise.context.control.ActivateRequestContext;
import jakarta.enterprise.inject.Instance;
import jakarta.inject.Inject;

/**
 * Using the configured database, Github installation records are tracked for all installations available for the
 * currently configured GH application. This is used over an in-memory cache as the calls to build the cache are long
 * running and have been known to fail sporadically.
 */
@ApplicationScoped
public class GithubInstallationUpdateTask {
    private static final Logger LOGGER = LoggerFactory.getLogger(GithubInstallationUpdateTask.class);

    @ConfigProperty(name = "eclipse.git-eca.tasks.gh-installation.enabled", defaultValue = "true")
    Instance<Boolean> isEnabled;
    
    @Inject
    GithubHelper helper;

    @PostConstruct
    void init() {
        // indicate to log whether enabled to reduce log spam
        LOGGER.info("Github installation DB cache task is{} enabled.", Boolean.TRUE.equals(isEnabled.get()) ? "" : " not");
    }

    /**
     * Every 1h, this method will query Github for installations for the currently configured application. These
     * installations will be translated into database entries for a persistent cache to be loaded on request.
     */
    @Scheduled(every = "1h", delay = 1)
    @ActivateRequestContext
    public void revalidate() {
        // if not enabled, don't process any potentially OOD records
        if (!Boolean.TRUE.equals(isEnabled.get())) {
            return;
        }
        helper.updateAppInstallationRecords();
    }
}
